! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: emanom
!
!
! Authors:  Kip Kuntz - NASA/GSFC/JHU
!           Brendan Perry - NASA/GSFC Perl->F95 conversion
!          
! Description: Determines whether EPIC MOS chips are in an anomalous state
! 
! Comments: New task for SAS-19 (2020)
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module emanom_mod

  use dal
  use param
  use gofutils_mod ! in goflib
  use dataFilter ! in selectlib
  use errorHandling
  use gofio_mod ! in goflib

  implicit none 
  
  private 

  public emanom

contains

  !---------------------------------------------------------------------
  subroutine emanom
  !---------------------------------------------------------------------

    type(tableT) :: inEVLI_T, outCORN_T
    type(DataSetT) :: inEVLI_S, outCORN_S
    type(ArrayT)   :: inEVLI_A
    character(len=1024) ::corndef, expression
    character(len=256) :: inEVLI_F, outCORN_F
    character(len=80) :: commstr ! comment string for fits file
    character(len=25) :: smode
    character(len=30) :: logfile
    character(len=10) :: obsid
    character(len=8) :: instrume, strcurrLT, hrstr, expid, &
         lBl, lBh, hBl, hBh
    character(len=6) :: filter
    character(len=2) :: inst
    character(len=1) :: strcurrchip, state
    integer :: loBandLo, loBandHi, hiBandLo, hiBandHi, nRows, frmtime, &
         elow, ehigh
    integer :: mode ! 1=read, 2=write, 3=create, 4=modify
    integer :: status=0, currchip, lun, ios, cornRows, typeF, revolut
    integer, dimension(7) :: anom_pos
    logical :: warndie, writekeys, writelog, keepcorner
    real(kind=single) :: back, area, expo
    real(kind=single) :: hard, high_band, low_band, sighard
    real(kind=single), dimension(7) :: anom_lim, noan_lim
    real(kind=double) :: currLT
 
    ! Some hard-wired parameters NOT dependent on the instrument
    ! Energy limits of bands for hardness ratio
    
    loBandLo = 400
    loBandHi = 800
    hiBandLo = 2500
    hiBandHi = 5000
    write(lBl,'(I5)') loBandLo
    write(lBh,'(I5)') loBandHi
    write(hBl,'(I5)') hiBandLo
    write(hBh,'(I5)') hiBandHi
    lBl = adjustl(lBl)
    lBh = adjustl(lBh)
    hBl = adjustl(hBl)
    hBh = adjustl(hBh)

    ! The first input paramter is the name of the input event file 
    ! to which pattern and flag filter has not been applied
    ! SPF filtering is OK

    inEVLI_F = stringParameter("eventfile")
    write(errstr,*) "Input event list:      ", trim(adjustl(inEVLI_F))
    call message(AppMsg,Always)
    ! The second input parameter is the name of the output corner file
    outCORN_F = stringParameter("cornerfile")
    write(errstr,*) "Output CCD corner file ", trim(adjustl(outCORN_F)) 
    call message(AppMsg,Always)
    writekeys  = booleanParameter("writekeys")  ! write keys to EVLI
    writelog   = booleanParameter("writelog")   ! write results to text log
    keepcorner = booleanParameter("keepcorner") ! retain corner data created below
    
    if (writekeys) then 
       mode = 4 ! will modify evli below. read/write/create/modify = 1/2/3/4
       status = 0
       warndie = .TRUE. ! die gracefully if no file
       call check_open_set(inEVLI_F,inEVLI_S,mode,status,warndie)
    else
       mode = 1 ! will modify evli below. read/write/create/modify = 1/2/3/4
       status = 0
       warndie = .TRUE. ! die gracefully if no file
       call check_open_set(inEVLI_F,inEVLI_S,mode,status,warndie)
    end if
    !
    ! inEVLI_T = table(inEVLI_S,"EVENTS") ! opened set above, now open table = EVENTS
    !
    ! Get critical header keywords 
    !
    status = 0
    warndie = .TRUE.
    typeF = 1 ! 1 = EVLI, 2 = SPEC, 3 = IMG

    call read_header_info(typeF,inEVLI_S,inEVLI_T,inEVLI_A,inst,obsid,expid,&
         smode,frmtime,filter,revolut,back,area,expo,elow,ehigh,nRows,warndie)
    
    if ( ( trim(adjustl(inst)) .NE. 'M1') .and. ( trim(adjustl(inst)) .NE. 'M2') ) then
       call error ("wrongInstrume",&
            "INSTRUME must be EMOS1/EMOS2 for anomalous state calculation")
    end if

    if ( ( trim(adjustl(filter)) .NE. 'Thick') .and. ( trim(adjustl(filter)) .NE. 'Medium') &
         .and. ( trim(adjustl(filter)) .NE. 'Thin1') .and. ( trim(adjustl(filter)) .NE. 'Closed')) then 
       call error ("wrongFilter",&
            "FILTER must be Thick/Medium/Thin1/Closed for anomalous state calculation")
    end if
    if ( trim(adjustl(filter)) .EQ. 'Closed')  then 
       call error ("CalClosed",&
            "emanom does not work with CalClosed data")
    end if
    if (trim(adjustl(inst)) .EQ. 'M1') then
       corndef="!(CIRCLE(100,-200,17700,DETX,DETY)||CIRCLE(834,135,17100,DETX,DETY)"&
            // "||CIRCLE(770,-803,17100,DETX,DETY)||BOX(-20,-17000,6500,500,0,DETX,DETY)"&
            // "||BOX(5880,-20500,7500,1500,10,DETX,DETY)||BOX(-5920,-20500,7500,1500,350,DETX,DETY)"&
            // "||BOX(-20,-20000,5500,500,0,DETX,DETY)||BOX(-12900,16000,250,4000,0,DETX,DETY)"&
            // "||BOX(80,18600,150,1300,0,DETX,DETY)||BOX(-10,-18800,125,1500,0,DETX,DETY))"
       anom_pos = (/0,0,0,1,1,0,0/)
       anom_lim = (/0.0,0.0,0.0,2.75,2.25,0.0,0.0/)
       noan_lim = (/0.0,0.0,0.0,3.00,3.60,0.0,0.0/)
    elseif (trim(adjustl(inst)) .EQ. 'M2') then
       corndef="!(CIRCLE(435,1006,17100,DETX,DETY)||CIRCLE(-34,68,17700,DETX,DETY)"&
            // "||BOX(-20,-17000,6500,500,0,DETX,DETY)||BOX(5880,-20500,7500,1500,10,DETX,DETY)"&
            // "||BOX(-5920,-20500,7500,1500,350,DETX,DETY)||BOX(-20,-20000,5500,500,0,DETX,DETY))"
       anom_pos = (/0,1,0,0,1,0,0/) ! 0 added at end bhp
       anom_lim = (/0.0,2.10,0.0,0.0,2.00,0.0,0.0/) ! 0 added at end bhp
       noan_lim = (/0.0,2.50,0.0,0.0,4.00,0.0,0.0/) ! 0 added at end bhp
    else ! should've been caught above, but meh.
       call error ("wrongInstrume",&
            "INSTRUME must be EMOS1/EMOS2 for anomalous state calculation.")
    end if
    ! create 'unique' (withing OBSID) logfile name
    logfile = "mos" // trim(adjustl(inst(2:2))) // trim(adjustl(expid)) // '-anom.log' ! mos1S001-anom.log
    write(errstr,*) "Output Anomalous log filename: ", trim(adjustl(logfile))
    call message(AppMsg,Noisy)
    write(errstr,*) "Creating corner-only events file: ", trim(adjustl(outCORN_F))
    call message(AppMsg,Noisy)
    ! Create the corner-only events file 
    ! first, clone the event list
    if (keepcorner) then 
       outCORN_S = clone(trim(adjustl(inEVLI_F)),trim(adjustl(outCORN_F)),MODIFY)
    else
       outCORN_S = clone(trim(adjustl(inEVLI_F)),trim(adjustl(outCORN_F)),TEMP)
    end if
    outCORN_T = table(outCORN_S,"EVENTS") 
    expression = "((FLAG & 0x766aa000) == 0)&&" // trim(adjustl(corndef))
    call initFilter(trim(adjustl(expression))) ! see SAS task 'selectlib' 
    cornRows = doFilter(outCORN_T)

    write(errstr,*) "Number of corner-only file rows = ", cornRows
    call message(AppMsg,Noisy)

    ! create output text log file (less generic name useful for Trend Data input)
    if (writelog) then 
       lun = GetLun() ! in gofutils
       ios = 0        
       write(errstr,*) "Opening output log file", trim(adjustl(logfile))
       call message(AppMsg,Noisy)
       open (lun,file=trim(adjustl(logfile)), iostat=ios)
       if (ios .ne. 0) then
          call warning ("noLUN","Cannot get LUN for log output file")
       end if
    end if
    !
    do currchip = 2, 7
       currLT = 0.
       write(strcurrchip,'(I1)') currchip
       strcurrLT = "LIVETI0" // trim(adjustl(strcurrchip))
       if (HasAttribute(inEVLI_T,trim(adjustl(strcurrLT)))) then
          currLT = real64Attribute(inEVLI_T,trim(adjustl(strcurrLT)))
          write(errstr,*) "Current chip: ", strcurrchip," Livetime: ",currLT
          call message(AppMsg,Noisy)
       else
          write(errstr,*) "Input evli file does not have ", strcurrLT," attribute"
          call message(AppMsg,Noisy)
       endif
       !       
       if (currLT .GT. 0.) then
          expression = ""
          ! doFilter (get rate in 0.4-0.8 keV band).
          expression = "(PATTERN<=12)&&((FLAG & 0x766aa000)==0)"&
               // "&&(CCDNR==" // trim(adjustl(strcurrchip)) // ")&&(PI in ["&
               // trim(adjustl(lBl)) // ":" // trim(adjustl(lBh)) // "])"
          call initFilter(trim(adjustl(expression))) ! in SAS task 'selectlib'
          write(errstr,*) "Selection expression, low_band = ", trim(adjustl(expression))
          call message(AppMsg,Noisy)
          low_band = doSelect(outCORN_T)
          ! doFilter (get rate in 2.5-5.0 keV band)
          expression = ""
          expression = "(PATTERN<=12)&&((FLAG & 0x766aa000)==0)"&
               // "&&(CCDNR==" // trim(adjustl(strcurrchip)) // ")&&(PI in ["&
               // trim(adjustl(hBl)) // ":" // trim(adjustl(hBh)) // "])"
          !          // "&&(CCDNR==" // trim(strcurrchip) // ")&&(PI in [2500:5000])"
          call initFilter(trim(adjustl(expression))) ! in SAS task 'selectlib'
          write(errstr,*) "Selection expression,  hi_band = ", trim(adjustl(expression))
          call message(AppMsg,Noisy)
          high_band = doSelect(outCORN_T)

          ! get the hardness ratio and the uncertainty of the hardness rate
          if (low_band .le. 0.) then ! catch divide by zero below
             hard = -999.
             sigHard = -999.
             state = "U" ! undetermined
          else
             hard = high_band / low_band
             sigHard = SQRT ((1. + high_band / low_band) * high_band / low_band / low_band )
             ! determine if an anomalous state

             if (anom_pos(currchip) .eq. 0.) then
                state = "G"
             else
                if (hard .lt. anom_lim(currchip)) then
                   state = "B"
                else
                   if (hard .gt. noan_lim(currchip)) then
                      state = "G"
                   else
                      state = "I"
                   end if
                end if
             end if
          end if          
       else
          hard = -999.
          sigHard = -999.
          state = "O"
       end if ! currLT > 0.

       ! write to the header of the in_EVLI

       if (writekeys) then 
          hrstr = "ANOMHR" // strcurrchip
          commstr = "Hardness Ratio for chip " // strcurrchip
          call setAttribute(inEVLI_S,trim(adjustl(hrstr)),hard,"",trim(adjustl(commstr)))
          hrstr = "ANOMHE" // strcurrchip
          commstr = "Uncer. for HR for chip " // strcurrchip
          call setAttribute(inEVLI_S,trim(adjustl(hrstr)),sigHard,"",trim(adjustl(commstr)))
          hrstr = "ANOMFL" // strcurrchip
          commstr = "Anomalous state for chip " // strcurrchip
          call setAttribute(inEVLI_S,trim(adjustl(hrstr)),state,trim(adjustl(commstr)))
          write(errstr,*) "Writing STATES to EVLI: ANOMHR: ",hard," ANOMHE: ",sigHard," ANOMFL: ",state
          call message(AppMsg,Noisy)
       end if

       write(errstr,*) trim(adjustl(inst))," CCD: ",strcurrchip," Hard = ",hard," +/- ",sigHard," Status: ",state
       call message(AppMsg,Always)
       if (writelog) write(lun,*)  &
            trim(adjustl(inst))," CCD: ",strcurrchip," Hard = ",hard," +/- ",sigHard," Status: ",state

    end do ! currchip = 2, 7
    call release(inEVLI_S)
    write(errstr,*) "Status: G is good at all energies"
    if (writelog) write(lun,*) "Status: G is good at all energies"
    call message(AppMsg,Always)
    write(errstr,*) "      : I is intermediate for E<1 keV"
    if (writelog) write(lun,*) "      : I is intermediate for E<1 keV"
    call message(AppMsg,Always)
    write(errstr,*) "      : B is bad for E<1 keV"
    if (writelog) write(lun,*) "      : B is bad for E<1 keV"
    call message(AppMsg,Always)
    write(errstr,*) "      : O is off, chip not in use"
    if (writelog) write(lun,*) "      : O is off, chip not in use"
    call message(AppMsg,Always)
    write(errstr,*) "      : U is undetermined (low band <= 0)"
    if (writelog) write(lun,*) "      : U is undetermined (low band counts <= 0)"
    call message(AppMsg,Always)
    ios = 0
    if (writelog) then
       close(lun,iostat=ios)
       if (ios .ne. 0) then
          write(errstr,*) "Could not close LUN of output log file: ",trim(adjustl(logfile))
          call warning ("noCLOSE",trim(adjustl(errstr)))
       end if
    end if
    if (keepcorner) then
       call release (outCORN_S)
    else
       call discardDataSet(trim(adjustl(outCORN_F)))
    endif
  end subroutine emanom

end module emanom_mod
